﻿/*
 * The Stoichiometer
 * @author mrhomescientist
 * @date 12/14/2017
 * 
 * Small tool to assist in stoichiometry calculations. The user inputs a chemical equation and the program outputs
 * the molecular weights of each molecule entered. Then, the user chooses a molecule and enters the desired mass,
 * and the program outputs the masses of all of the other molecules in the equation.
 * This program does NOT balance equations, so they must be properly balanced upon entry to give useful results.
 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Stoichiometer
{
    public partial class Form1 : Form
    {
        //dictionary of elements and formula weights
        Dictionary<string, double> d = new Dictionary<string, double>();
        String[] element = { "H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", "Sc",
                               "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", "Nb",
                               "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm",
                               "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl",
                               "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md",
                               "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og" };
        double[] mass = { 1.008, 4.003, 6.941, 9.012, 10.811, 12.011, 14.007, 15.999, 18.998, 20.180, 22.990, 24.305, 26.982, 28.086, 30.974, 32.065,
                            35.453, 39.948, 39.098, 40.078, 44.956, 47.867, 50.942, 51.996, 54.938, 55.845, 58.933, 58.693, 63.546, 65.39, 69.723, 72.64,
                            74.922, 78.96, 79.904, 83.80, 85.468, 87.62, 88.906, 91.224, 92.906, 95.94, 98, 101.07, 102.91, 106.42, 107.87, 112.41, 114.82,
                            118.71, 121.76, 127.60, 126.90, 131.29, 132.91, 137.33, 138.91, 140.12, 140.91, 144.24, 145, 150.36, 151.96, 157.25, 158.93,
                            162.50, 164.93, 167.26, 168.93, 173.04, 174.97, 178.49, 180.95, 183.84, 186.21, 190.23, 192.22, 195.08, 196.97, 200.59, 204.38,
                            207.2, 208.98, 209, 210, 222, 223, 226, 227, 232.04, 231.04, 238.03, 237, 244, 243, 247, 247, 251, 252, 257, 258, 259, 266, 267,
                            268, 269, 270, 277, 278, 281, 282, 285, 286, 289, 290, 293, 294, 294 };
        //other global arrays
        string[] equation;
        double[] masses;
        int[] coefficients;
        TextBox[] boxes;


        public Form1()
        {
            InitializeComponent();

            //populate dictionary
            for (int i = 0; i < element.Length; i++)
                d.Add(element[i], mass[i]);
        }


        private void runButton_Click(object sender, EventArgs e)
        {
            //**********************
            //clear old data
            //**********************
            panel1.Controls.Clear();
            
            //**********************
            //fill reactant & product arrays from GUI
            //**********************
            //split along + and = chars & trim whitespace
            equation = Array.ConvertAll(eqBox.Text.Split(new char[] {'=', '+'}), p => p.Trim());
            coefficients = new int[equation.Length];
            masses = new double[equation.Length];

            //**********************
            //determine molecular weight of each entry in each array
            //**********************
            try
            {
                for (int i = 0; i < equation.Length; i++)
                    masses[i] = molecularMass(equation[i], i, true);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            //**********************
            //build lower portion of the GUI
            //**********************
            //write equation labels
            string[] eq = eqBox.Text.Split(' ');
            for (int i = 0; i < eq.Length; i++)
            {
                Label label = new Label();
                label.Location = new Point(100 * i, 0);
                label.Font = new Font("Microsoft Sans Serif", 12F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
                label.AutoSize = true;
                label.Text = eq[i];
                panel1.Controls.Add(label);
            }
            //write molecular weights
            for (int i = 0; i < equation.Length; i++)
                {
                TextBox box = new TextBox();
                box.Location = new Point(100 * (2*i), 30);  //23 is the height of a label with 12pt font
                box.Font = new Font("Microsoft Sans Serif", 12F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
                box.AutoSize = true;
                box.ReadOnly = true;
                box.Text = "" + masses[i];
                panel1.Controls.Add(box);

                Label label = new Label();
                label.Location = new Point(100 * (2*i+1), 30);  //23 is the height of a label with 12pt font
                label.Font = new Font("Microsoft Sans Serif", 12F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
                label.AutoSize = true;
                label.Text = "g/mol";
                panel1.Controls.Add(label);
            }
            //add stoichiometry text boxes
            boxes = new TextBox[equation.Length];
            for (int i = 0; i < boxes.Length; i++)
            {
                var txt = new TextBox();
                txt.Name = "box" + i;
                txt.Location = new Point(100 * (2 * i), 60);
                txt.Font = new Font("Microsoft Sans Serif", 14F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
                txt.AutoSize = true;
                txt.KeyUp += new KeyEventHandler(tb_KeyUp);
                boxes[i] = txt;
                panel1.Controls.Add(txt);

                Label label = new Label();
                label.Location = new Point(100 * (2 * i + 1), 60);  //23 is the height of a label with 12pt font
                label.Font = new Font("Microsoft Sans Serif", 12F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
                label.AutoSize = true;
                label.Text = "g";
                panel1.Controls.Add(label);
            }
            //reveal final label
            label2.Visible = true;
        }


        private void tb_KeyUp(object sender, KeyEventArgs e)
        {
             //enter key was just pressed
            if (e.KeyCode == Keys.Enter)
            {
                //get info from the current textbox
                TextBox box = (TextBox)sender;
                int boxID = int.Parse(box.Name.Substring(3));
                double mass = Double.Parse(box.Text);
                //stoichiometerize!
                //perform first half of calculation only once
                double stoich = mass / masses[boxID] / coefficients[boxID];
                for (int i = 0; i < boxes.Length; i++)
                {
                    //calculate second half and display up to 3 decimals
                    if (i != boxID)
                        boxes[i].Text = (stoich * coefficients[i] * masses[i]).ToString("#.000");
                }
            }
        }


        /*********************
         * Helper method to process each item character-by-character, to determine total molecular weight of the item
         *********************/
        private double molecularMass(string item, int index, bool first)
        {
	        double mass = 0;
	        double lastMass = 0;
	        int j = 0;
	        string str;
	        //check first chars for a coefficient (if this isn't a recursive call)
            int num, temp;
            if (first)
            {
	            temp = 1;
	            for (int l = 1; l < item.Length; l++)
	            {
		            str  = item.Substring(0,l);
		            if (!int.TryParse(str, out num))
			            break;
                    temp = num;
		            j++;
	            }
                coefficients[index] = temp;
           }

	        //continue processing this item
	        for (;  j < item.Length; j++)
	        {
		        str = item.Substring(j,1);
		        //if char is a parenthesis, create a substring and recursively call this method on the substring
		        if (str == "(")
		        {
			        int k = item.IndexOf(")");
                    lastMass = molecularMass(item.Substring((j + 1), (k - 1 - j)), j, false);
			        mass += lastMass;
			        j = k;
		        }
		        else
		        {
			        //determine if char is a number
			        //ONLY WORKS FOR 1 DIGIT NUMBERS
                    if (int.TryParse(str, out num))
                    {
                        temp = num;
                        //check for more digits
                        if (j + 1 < item.Length)
                        {
                            int k = j + 1;
                            for (int l = 1; (l + j) < item.Length; l++)
                            {
                                str = item.Substring(k, l);
                                if (!int.TryParse(str, out num))
                                    break;
                                temp = int.Parse("" + temp + num);
                                j++;
                            }
                        }
                        mass += lastMass * (temp - 1);
                    }
                    //make sure we aren't at the last character
                    else if (j + 1 < item.Length)
                    {
                        //check if next char is uppercase or a number - if so, this is a 1-char element
                        if (char.IsUpper(item[j + 1]) || int.TryParse(item.Substring(j + 1, 1), out num))
                        {
                            if (d.ContainsKey(str))
                            {
                                lastMass = d[str];
                                mass += lastMass;
                            }
                            else
                            {
                                //ERROR: not found!
                                throw new KeyNotFoundException("Element " + str + " not found! Make sure to use proper capitalization!");
                            }
                        }
                        else
                        {
                            //else, this is a 2-char element
                            str += item.Substring(j + 1, 1);
                            if (d.ContainsKey(str))
                            {
                                lastMass = d[str];
                                mass += lastMass;
                                j++;
                            }
                            else
                            {
                                //ERROR: not found!
                                throw new KeyNotFoundException("Element " + str + " not found! Make sure to use proper capitalization!");
                            }
                        }

                    }
                    else
                    {
                        //check last char for a 1-char element
                        if (d.ContainsKey(str))
                        {
                            lastMass = d[str];
                            mass += lastMass;
                        }
                        else
                        {
                            //ERROR: not found!
                            throw new KeyNotFoundException("Element " + str + " not found! Make sure to use proper capitalization!");
                        }
                    }
		        }
	        }
	        return mass;
        }
    }
}
